#! /usr/bin/julia

# This is an implementation of the divide & conquer algorithm for
# computing the closest pair of points in a set of n points in
# O(n log n) time. Written by Sariel Har-Peled.

using LinearAlgebra

Point = Tuple{Float64, Float64};
Points = Vector{Point};

mutable struct  Solution
    d::Float64
    i::Int64
    j::Int64
end
function  Sol_set( sol::Solution, _d, _i, _j ) 
    sol.d, sol.i, sol.j = _d, _i, _j;
end

struct  PointSet
    P::Points;
    ord_x::Vector{Int64};
    ord_y::Vector{Int64};
end

len( PS::PointSet ) = length( PS.ord_x );    # number of points in PS
d( p, q ) =  norm( ( p[1] - q[1], p[2] - q[2] ) ); # Distance between points
PS_ord_y( Q, l ) = (Q.P[ Q.ord_y[ l ] ])[ 2 ];

struct SortByCoord <: Base.Order.Ordering
    P::Points;
    coord::Int64
end

import Base.Order.lt
function  lt(o::SortByCoord, a, b)
    p,q = o.P[ a ], o.P[ b ];
    return  (p[o.coord] < q[o.coord] );
end

function  PointSet( _P::Points )
    PS = PointSet( _P, collect( 1:length( _P ) ), collect( 1:length( _P ) ) );

    # Sort the points by x-axis, and store the ordering in PS.ord_x
    sort!( PS.ord_x, order=SortByCoord( PS.P, 1 ) );

    # Sort the points by y-axis, and store the ordering in PS.ord_y
    sort!( PS.ord_y, order=SortByCoord( PS.P, 2 ) );

    return  PS;
end

# Computing subset of points in _PS that fulfill the condition f.
function  PS_filter( P::PointSet, f )
    return    PointSet( P.P, filter( i -> f( P.P[ i ] ), P.ord_x ),
                             filter( i -> f( P.P[ i ] ), P.ord_y ) );
end

# Use the elevator algorithm to discover any closest pair that is
# better than the current solution.
function  nn_middle( VPS::PointSet, mid, sol::Solution )
    low, hi = mid - sol.d, mid + sol.d
    P = PS_filter( VPS, p -> ( low <= p[1] <= hi ) );
    ( len( P ) < 2 )  &&  return  sol;

    b = t = 1;  # Bottom/top of elevator
    while  ( t <= len( P ) )
        # Delete bottom point ∈ elevator if irrelevant
        while  ( b < t ) && ( PS_ord_y( P, b ) < (PS_ord_y( P, t ) - sol.d) )
            b = b + 1;
        end

        # Find nn in points in the elevator to the top point in it.
        # The elevator points are all points with indices in
        # PS.ord_y[ b:t ]
        i_p = P.ord_y[ t ];
        p = P.P[ i_p ];

        for  j ∈ b:(t-1)
            i_q = P.ord_y[ j ]
            q = P.P[ i_q ]
            if  ( d( p, q ) < sol.d )
                Sol_set( sol, d( p, q ), i_p, i_q );
            end
        end
        t = t + 1;
    end

    return  sol;
end

function  closest_pair_inner( PS::PointSet, sol )
    ( len( PS ) < 2)  &&  return  sol;

    mid = PS.ord_x[ len( PS )  ÷  2 ];
    x_mid = PS.P[ mid ][ 1 ];

    # A subtlety: All the points with x-coordinat x_mid, would be
    # checked by the call nn_middle, so no need to include them in the
    # two recursive calls.

    # Dividing...
    PS_L = PS_filter( PS, p -> ( p[1] < x_mid ) );  # Left points
    PS_R = PS_filter( PS, p -> ( p[1] > x_mid ) );  # Right points

    # Recursing
    closest_pair_inner( PS_L, sol ); # recurse left
    closest_pair_inner( PS_R, sol ); # recurse right

    # Conquering
    nn_middle( PS, x_mid, sol );

    # declaring victory, and withdrawing
    return  sol;
end

function  closest_pair( PS::PointSet )
    sol = Solution( d(PS.P[1], PS.P[2]), 1, 2 );
    return  closest_pair_inner( PS, sol );
end

function  closest_pair_brute_force( PS::PointSet )
    sol = Solution( d(PS.P[1], PS.P[2]), 1, 2 );

    for  i in 1:length( PS.ord_x )-1
        i_r = PS.ord_x[ i ];
        p = PS.P[ i_r ];
        for  j in i+1:length( PS.ord_x )
            j_r =  PS.ord_x[ j ]
            q = PS.P[ j_r ];
            if  ( d( p, q ) < sol.d )
                Sol_set( sol,  d(PS.P[i_r], PS.P[j_r]), i_r, j_r );
            end
        end
    end

    return  sol;
end

function  main()
    # The D&C algorithm seems faster than the naive algorithm around n ≈ 200
    n = 200;

    println( "n : ", n );
    PS = PointSet(  [ (rand(), rand()) for i ∈ 1:n] );

    @time sol = closest_pair( PS );
    @time sol_bf = closest_pair_brute_force( PS );

    println( "cp    : ", sol );
    println( "cp_bf : ", sol_bf );
end

main();
